// HC11/CS8900 ethernet IP stack
//
// $Id: ip.c,v 2.1 2001/12/25 23:08:47 tom Exp tom $
//
// Copyright (C) 2000-1 Tom Parker (tom-hc11@wiresncode.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Author's note: This is free software, for commerical use I would
// appreciate some recognition though.  If you are using this and like
// it send me some email or a postcard.  Code changes for inclusion
// are also welcome.
//
// Tom Parker, PO Box 2165, Fitzroy, VIC, 3065, AUSTRALIA (tom-hc11@wiresncode.com)
//

#include "ether.h"
#include "udp.h"
#include "arp.h"
#include "icmp.h"
#include "tcp.h"

#include "serial.h"
#include "cs8900.h"

//-----------------------------------------------------------------
// Globals
u_int8_t ourIP[sizeof(struct in_addr)];
u_int8_t ourBCAST[sizeof(struct in_addr)];
volatile u_int8_t bootpPending;
volatile u_int8_t bootpTimeout;


//-----------------------------------------------------------------
void
ip_init()
{
    // initial IP broadcast address
   memset(ourBCAST, 0xff, sizeof(struct in_addr));
#ifdef BOOTP
   bootpPending = 1;
#else
   bootpPending = 0;
#endif
   bootpTimeout = 0;
}

//-----------------------------------------------------------------
// send an ip packet (protocol, length & destination address are
// assumed to be filled out).  Packet is MAC size packet ready to go.
void
ip_send()
{
   struct ether_header* ethHdr = (struct ether_header*) txBuffer;
   struct ip* ipPkt = (struct ip*) (txBuffer + sizeof(struct ether_header));
   const u_int16_t* mem = (u_int16_t*) ipPkt;
   u_int32_t newcksum = 0;
   register int i = 0;

   serial_print_debug('i');

   ipPkt->ip_v = IPVERSION;
   ipPkt->ip_hl = 5;
   ipPkt->ip_tos = 0;
   ipPkt->ip_off = 0;

   // check bootp status
   if (bootpPending == 0)
   {
      u_int8_t arpRetryCount = 0;

      serial_print_debug('1');

      // address is assigned so use our IP & a specific MAC address
      memcpy(&ipPkt->ip_src, ourIP, sizeof(struct in_addr));

      if (memcmp(&ipPkt->ip_dst, arpIP, sizeof(struct in_addr)))
      {
         arpValid = 0;
      }

      arpTimeout = 0;
      while(arpValid != 1)
      {
         if (arpTimeout == 0)
         {
            if (arpRetryCount++ > MAXARPATTEMPTS)
            {
               serial_print_string("Aborting TX - ARP failed\r\n");
               
               return;
            }

            arp_send(ARPOP_REQUEST, 0, (u_int8_t *) &ipPkt->ip_dst);
            arpTimeout = ARPINTERVAL;
         }

         cs8900_poll();
      }
      memcpy(ethHdr->ether_dhost, arpMAC, ETHER_ADDR_LEN);
   }
   else
   {
      // use null IP address & bcast MAC address if bootp is not finished

      serial_print_debug('2');
      memset(ethHdr->ether_dhost, 0xff, ETHER_ADDR_LEN);
      bzero((void *) &ipPkt->ip_src, sizeof(struct in_addr)); 
   }

   // checksum
   ipPkt->ip_sum = 0;
   for(i = 0; i < 10; i++, mem++)
   {
      newcksum += *mem;
   }
   ipPkt->ip_sum = htons((u_int16_t) ~((newcksum >> 16) + (newcksum & 0xffff)));

   // MAC
   ethHdr->ether_type = ETHERTYPE_IP;

   cs8900_send(txBuffer, ntohs(ipPkt->ip_len) + sizeof(struct ether_header));
}

//-----------------------------------------------------------------
// IP packet receive & action
void
ip_receive()
{
   const struct ip* ipPkt = (struct ip*) (rxBuffer + sizeof(struct ether_header));

   serial_print_debug('I');

   // extract IPv4, non fragmented, optionless packets with a valid length
   // destined for us (if bootp is done) or broadcast
   if ((ipPkt->ip_v == IPVERSION && 
	(ipPkt->ip_off & IP_OFFMASK) == 0 && 
	ntohs(ipPkt->ip_len) > 20 &&
	ipPkt->ip_hl == 5) &&
       ((!memcmp(ourIP, &ipPkt->ip_dst, sizeof(struct in_addr)) && bootpPending == 0) ||
	(!memcmp(ourBCAST, &ipPkt->ip_dst, sizeof(struct in_addr)))))
   {
      const struct ether_header* ethHdr = (struct ether_header*) rxBuffer;
      u_int32_t newcksum = 0;
      u_int16_t* mem = (u_int16_t*) ipPkt;
      register u_int16_t i = 0;

      // save MAC address and IP address to prevent ARP
      memcpy(arpMAC, ethHdr->ether_shost, ETHER_ADDR_LEN);
      memcpy(arpIP, &ipPkt->ip_src, sizeof(struct in_addr));
      arpValid = 1;

      serial_print_debug('1');
      
      // checksum
      for (i = 0; i < 10; i++, mem++)
      {
	 // skip existing checksum
	 if (i != 5)
	 {
	    newcksum += *mem;
	 }
      }
      newcksum = ~((newcksum >> 16) + (newcksum & 0xffff));
      
      if ((newcksum & 0xffff) == ipPkt->ip_sum)
      {
	 serial_print_debug('2');

	 // Valid packet
	 switch(ipPkt->ip_p)
	 {
#ifdef ICMP
	    case IPPROTO_ICMP:
	    {
	       icmp_receive(ntohs(ipPkt->ip_len) - 20);
	       break;
	    }
#endif
#ifdef UDP
	    case IPPROTO_UDP:
	    {
	       udp_receive(ntohs(ipPkt->ip_len) - 20);
	       break;
	    }
#endif
#ifdef TCP
            case IPPROTO_TCP:
            {
               tcp_receive(ntohs(ipPkt->ip_len) - 20);
               break;
            }
#endif
	 }
      }
      else
      {
	 serial_print_string("IP checksum error 0x");
	 serial_print_hex_int(newcksum,1);
	 serial_print_string(" != 0x");
	 serial_print_hex_int(ipPkt->ip_sum,1);
	 serial_print_crlf();
      }
   }
   
}

