// HC11/CS8900 ethernet IP stack
//
// $Id: udp.c,v 1.2 2001/12/25 23:07:27 tom Exp tom $
//
// Copyright (C) 2000-1 Tom Parker (tom-hc11@wiresncode.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// Author's note: This is free software, for commerical use I would
// appreciate some recognition though.  If you are using this and like
// it send me some email or a postcard.  Code changes for inclusion
// are also welcome.
//
// Tom Parker, PO Box 2165, Fitzroy, VIC, 3065, AUSTRALIA (tom-hc11@wiresncode.com)
//

#include "ether.h"

#include "ntp.h"
#include "bootp.h"
#include "ip.h"
#include "serial.h"
#include "cs8900.h"
#include "timer.h"

//-----------------------------------------------------------------
// send a udp packet given the address, length & payload in txBuffer
void
udp_send(const u_int16_t aSourcePort, 
	 const u_int16_t aDestPort, 
	 const u_int8_t* aAddress, 
	 const u_int16_t aLength)
{
   struct ip* ipPkt = (struct ip*) (txBuffer + sizeof(struct ether_header));
   struct udphdr* udpPkt = (struct udphdr*) (txBuffer + sizeof(struct ether_header) + sizeof(struct ip));
   const u_int16_t* mem = (u_int16_t*) &(ipPkt->ip_src);
   u_int32_t newcksum = IPPROTO_UDP + sizeof(struct udphdr) + aLength;
   register u_int16_t i = 0;
   u_int16_t length = aLength;

   serial_print_debug('u');

   // UDP header
   udpPkt->source = htons(aSourcePort);
   udpPkt->dest = htons(aDestPort);
   udpPkt->len = htons(length + sizeof(struct udphdr));
   udpPkt->check = 0;

   // destination IP
   memcpy(&ipPkt->ip_dst.s_addr, aAddress, sizeof(struct in_addr));

   // fill source address for checksum - use a zero ip address if we
   // do not have one
   if (bootpPending == 0)
   {
      memcpy(&ipPkt->ip_src, ourIP, sizeof(struct in_addr));
   }
   else
   {
      bzero((void *) &ipPkt->ip_src, sizeof(struct in_addr)); 
   }
   
   // packet length, protocol & id are the same
   ipPkt->ip_len = htons(length + sizeof(struct ip) + sizeof(struct udphdr));
   ipPkt->ip_p = IPPROTO_UDP;
   ipPkt->ip_id = 0;
   ipPkt->ip_ttl = 64;

   // checksum includes protocol (see above), udp length (see
   // above) and IP source & destination
   newcksum += ntohs(*mem++);
   newcksum += ntohs(*mem++);
   newcksum += ntohs(*mem++);
   newcksum += ntohs(*mem);

   // packet checksum
   mem = (u_int16_t*) udpPkt;
   
   // pad odd packets
   if (length % 2)
   {
      *((u_int8_t *) mem + length) = 0;
   }

   // calc checksum (convert length to words & round odd bytes up)
   length = (length + 1) >> 1;
   for (i = 0; i < length; i++, mem++)
   {
      newcksum += ntohs(*mem);
   }

   // insert checksum & send
   udpPkt->check = htons((u_int16_t) ~((newcksum >> 16) + (newcksum & 0xffff)));
   ip_send();
}

//-----------------------------------------------------------------
// send a bootp packet given the address, length & payload in txBuffer
void
udp_bootp_send()
{
   struct bootp* bootpPkt = (struct bootp*) (txBuffer + sizeof(struct ether_header) + sizeof(struct ip) + sizeof(struct udphdr));
   u_int8_t destAddr[sizeof(struct in_addr)] = {0xff, 0xff, 0xff, 0xff};
   const u_int8_t rfc1048cookie[sizeof(struct in_addr)] =  VM_RFC1048;

   serial_print_debug('b');

   // fill packet inc. our hardware address
   bzero((void *) bootpPkt, sizeof(struct bootp));
   bootpPkt->bp_op = BOOTREQUEST;
   bootpPkt->bp_htype = HTYPE_ETHERNET;
   bootpPkt->bp_hlen = ETHER_ADDR_LEN;
   bootpPkt->bp_flags = htons(0x8000); // broadcast
   memcpy(bootpPkt->bp_vend, rfc1048cookie, sizeof(struct in_addr));
   memcpy((void *) bootpPkt->bp_chaddr, (void *) ourMAC, ETHER_ADDR_LEN);

   bootpTimeout = 0;
   cpu_serial_send('.');
   
   udp_send(IPPORT_BOOTPC, IPPORT_BOOTPS, destAddr, sizeof(struct bootp));
}

//-----------------------------------------------------------------
// UDP packet receive & action
void
udp_ntp_receive(const u_int16_t aLength)
{
   serial_print_debug('N');

   if (aLength >= sizeof(struct ntp))
   {
      struct ntp* ntpPkt = (struct ntp*) (rxBuffer + sizeof(struct ether_header) + sizeof(struct ip) + sizeof(struct udphdr));

      currentTime = ntpPkt->xmt.l_ui;
   }
   else
   {
      serial_print_string("\r\nUndersize NTP packet received ");
      serial_print_int(aLength);
      serial_print_string(" < ");
      serial_print_int(sizeof(struct ntp));
      serial_print_crlf();
   }
}

//-----------------------------------------------------------------
// BOOTP packet receive & action
void
udp_bootp_receive(const u_int16_t aLength)
{
   struct bootp* bootpPkt = (struct bootp*) (rxBuffer + sizeof(struct ether_header) + sizeof(struct ip) + sizeof(struct udphdr));

   serial_print_debug('B');

   // only interested in responses
   if (bootpPkt->bp_op == BOOTREPLY && 
       bootpPkt->bp_htype == HTYPE_ETHERNET &&
       bootpPending == 1 &&
       !memcmp((void *) bootpPkt->bp_chaddr, (void *) ourMAC, ETHER_ADDR_LEN))
   {
      const u_int8_t rfc1048cookie[sizeof(struct in_addr)] =  VM_RFC1048;

      // get our IP address
      memcpy(ourIP, &bootpPkt->bp_yiaddr.s_addr, sizeof(struct in_addr));

      serial_print_string("\r\nGot bootp response - addr=");
      serial_print_int(((u_int8_t *)&bootpPkt->bp_yiaddr.s_addr)[0]);
      cpu_serial_send('.');
      serial_print_int(((u_int8_t *)&bootpPkt->bp_yiaddr.s_addr)[1]);
      cpu_serial_send('.');
      serial_print_int(((u_int8_t *)&bootpPkt->bp_yiaddr.s_addr)[2]);
      cpu_serial_send('.');
      serial_print_int(((u_int8_t *)&bootpPkt->bp_yiaddr.s_addr)[3]);
      
      // check for options
      if (!memcmp(bootpPkt->bp_vend, rfc1048cookie, sizeof(struct in_addr)))
      {
	 u_int8_t* mem = bootpPkt->bp_vend + sizeof(struct in_addr);

	 while(*mem !=  TAG_END)
	 {
	    switch(*mem)
	    {
	       case TAG_IP_BROADCAST:
		  if (*(mem + 1) == 4)
		  {
		     // get broadcast address
		     memcpy(ourBCAST, mem + 2, sizeof(struct in_addr));

		     serial_print_string(", broadcast=");
		     serial_print_int(*(mem + 2));
		     cpu_serial_send('.');
		     serial_print_int(*(mem + 3));
		     cpu_serial_send('.');
		     serial_print_int(*(mem + 4));
		     cpu_serial_send('.');
		     serial_print_int(*(mem + 5));
		  }
		  break;
	       default:
		  // ignore for now
	    }

	    // skip TAG + length fields + option length
	    mem += *(mem + 1) + 2;
	 }
      }
      serial_print_crlf();

      // release flag
      bootpPending--;
   }
}

//-----------------------------------------------------------------
// UDP packet receive & action
void
udp_receive(const u_int16_t aLength)
{
   struct udphdr* udpPkt = (struct udphdr*) (rxBuffer + sizeof(struct ether_header) + sizeof(struct ip));
   u_int16_t length = aLength;

   serial_print_debug('U');

   if (length == ntohs(udpPkt->len))
   {
      const struct ip* ipPkt = (struct ip*) (rxBuffer + sizeof(struct ether_header));
      const u_int16_t* mem = (u_int16_t*) &(ipPkt->ip_src);
      u_int32_t newcksum = IPPROTO_UDP + length;
      register u_int16_t i = 0;

      // checksum includes protocol (see above), udp length (see
      // above) and IP source & destination
      newcksum += ntohs(*mem++);
      newcksum += ntohs(*mem++);
      newcksum += ntohs(*mem++);
      newcksum += ntohs(*mem);

      // packet checksum
      mem = (u_int16_t*) udpPkt;

      // pad odd packets
      if (length % 2)
      {
	 *((u_int8_t *) mem + length) = 0;
      }

      // calc checksum (convert length to words & round odd bytes up)
      length = (length + 1) >> 1;
      for (i = 0; i < length; i++, mem++)
      {
	 // skip existing checksum
	 if (i != 3)
	 {
	    newcksum += ntohs(*mem);
	 }
      }
   
      newcksum = ~((newcksum >> 16) + (newcksum & 0xffff));
      if ((newcksum & 0xffff) == ntohs(udpPkt->check))
      {
	 switch(ntohs(udpPkt->dest))
	 {
	    case 123: // NTP
	       serial_print_debug('1');
	       udp_ntp_receive(ntohs(udpPkt->len));
	       break;
	    case IPPORT_BOOTPC:
	       serial_print_debug('2');
	       udp_bootp_receive(ntohs(udpPkt->len));
	       break;
	 }
      }
      else
      {
	 serial_print_string("\r\nUDP checksum error 0x");
	 serial_print_hex_int(newcksum,1);
	 serial_print_string(" != 0x");
	 serial_print_hex_int(ntohs(udpPkt->check), 1);
	 serial_print_crlf();
	 cs8900_dump(rxBuffer, BUFFERLEN);
      }
   }
   else
   {
      serial_print_string("\r\nUDP length mismatch ");
      serial_print_int(length);
      serial_print_string(" != ");
      serial_print_int(udpPkt->len);
      serial_print_crlf();
   }
}

